/*
 * getent.c -- build local user tree for finger
 * It all started here.. (November 1985)
 *
 * Copyright (C) 1986, 1996  Philip L. Budne
 *
 * This file is part of "Phil's Finger Program".
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 */

# ifndef lint
static char *rcsid = "$Id: getent.c,v 3.9 2000/02/18 06:04:06 phil Exp $";
# endif /* lint not defined */

# include <sys/types.h>
# include <stdio.h>
# include <ctype.h>
# include "person.h"
# include "args.h"			/* for luser.h */
# include "luser.h"
# include "tsel.h"
# include "finger.h"

# ifdef HAVE_UTMPX
# include <utmpx.h>
# define utmp utmpx
# define getutent getutxent
# define endutent endutxent
# define ut_time ut_tv.tv_sec
# else  /* HAVE_UTMPX not defined */
# include <utmp.h>
# endif /* HAVE_UTMPX not defined */

extern struct utmp *getutent();		/* from utmp.c */
extern LUSER *treefind();		/* from getperson.c */

FORWARD GLOBAL LUSER *linsert();
FORWARD LOCAL  LUSER *mkluser();

LOCAL TSEL *first_tsel, *last_tsel;

# define LINESIZE sizeof( ((struct utmp *)0)->ut_line )	/* size of ut_line */

/* .see gtname() in output.c */
# define TTY_PREFIX "tty"
# define TTY_PREFIX_SIZE (sizeof(TTY_PREFIX)-1)

GLOBAL void init_tsel() {
    first_tsel = last_tsel = NULL;
}

GLOBAL void addtty( s )
    char *s;
{
    register TSEL *ts;
# ifdef MATCH_TTY_PREFIXES
    register l, l2;
# endif /* MATCH_TTY_PREFIXES defined */

    if( (ts = (TSEL *) malloc( sizeof( TSEL ) )) == NULL ) {
	perror("could not allocate tty selector");
	exit( 1 );
    }

    ts->ts_name = s;

# ifdef MATCH_TTY_PREFIXES
    l = l2 = strlen( s );
    if( l > LINESIZE ) {
	l = LINESIZE;
	if( l > LINESIZE - TTY_PREFIX_SIZE )
	    l2 = l - TTY_PREFIX_SIZE;
    }

    ts->ts_len = l;
    ts->ts_len2 = l2;
# endif /* MATCH_TTY_PREFIXES defined */
    ts->ts_next = NULL;
    if( first_tsel == NULL )
	first_tsel = last_tsel = ts;
    else {
	last_tsel->ts_next = ts;
	last_tsel = ts;
    }
}

LOCAL int hideut(ut)
    register struct utmp *ut;
{
    /* hide login sessions (WISH: use ut_pid to find process to display) */
# ifdef AIX_RS6K
    if (strcmp(ut->ut_line, ".") == 0)	/* AIX4 w/ CDE? */
	return TRUE;
# endif /* AIX_RS6K defined */
# ifdef __osf__
    if (strcmp(ut->ut_line, ":0") == 0)
	return TRUE;
# endif /* __osf__ defined */
# ifdef CDE
    if (strncmp(ut->ut_line, "dtremote", 8) == 0)
	return TRUE;
# endif /* CDE defined */
    return FALSE;
}

GLOBAL LTREE *maketree() {		/* create tree of users */
    register struct utmp *ut;
    LTREE *tree;
    LUSER *new;

    tree = NULL;

    while( (ut = getutent()) != NULL ) {

	if( ut->ut_name[0] == EOS )	/* entry has name? */
 	    continue;			/* no, get next */

	if( hideut(ut) )		/* hide login sessions */
	    continue;

	if( first_tsel != NULL ) {
	    register TSEL *t;
	    int ok;

	    ok = FALSE;
	    for( t = first_tsel; t != NULL; t = t->ts_next ) {
# ifdef MATCH_TTY_PREFIXES
		/* old behavior; +xxx match xxx* or ttyxxx* */
		if( strncmp( ut->ut_line, t->ts_name, t->ts_len ) == 0 ||
		   (strncmp( ut->ut_line, TTY_PREFIX, TTY_PREFIX_SIZE ) == 0 &&
		    strncmp( ut->ut_line + TTY_PREFIX_SIZE,
			    t->ts_name, t->ts_len2 ) == 0) ){
		    ok = TRUE;
		    break;
		} /* if strncmp */
# else  /* MATCH_TTY_PREFIXES not defined */
		/*
 		 * new behavior; +xxx matches only xxx or ttyxxx since
		 * SVR4 has {pts,term}/{n,nn,nnn}, "." was making a mess!
		 */
		if (strcmp(ut->ut_line, t->ts_name) == 0 ||
		    (strncmp(ut->ut_line, TTY_PREFIX, TTY_PREFIX_SIZE) == 0 &&
		     strcmp(ut->ut_line + TTY_PREFIX_SIZE, t->ts_name) == 0)) {
		    ok = TRUE;
		    break;
		} /* if strcmp... */
# endif /* MATCH_TTY_PREFIXES not defined */
	    } /* for */
	    if( !ok )
		continue;
	} /* first_tsel */

	new = mkluser(ut);
	tree = linsert(new, tree);
    }
    endutent();
    return( tree );
} /* maketree */

GLOBAL LTREE *linsert(new, tree) 	/* insert new user into tree */
register LUSER *new;
register LTREE *tree;
{
    if( tree == NULL )			/* empty tree. return new entry */
        return( new );

    if( strcmp(new->u_user, tree->u_user) < 0 ) /* less than root? */
	 tree->u_left = (LUSER *)linsert(new, tree->u_left); /* put on left */
    else
         tree->u_right = (LUSER *)linsert(new, tree->u_right);/* put on right */
					/* dupes are added to right */
					/* (.see pwtree) because we assume */
					/* that the utmp file is in tty name */
					/* order */

    return( tree );			/* return the tree */
}

GLOBAL LUSER *newluser() {		/* create empty luser struct */
    register LUSER *new;

    if( (new = (LUSER *)malloc( sizeof( LUSER ) )) == NULL ) {
	fprintf(stderr, "malloc failed in newluser\n");
	exit( 1 );
    }
    new->u_person = NULL;
    new->u_left = new->u_right = NULL;
    new->u_flags = 0;
    new->u_sw = Sw;			/* get per-person switches */
    return( new );
} /* newluser */

LOCAL LUSER *mkluser( ut )		/* create a luser from utmp */
register struct utmp *ut;
{
    register LUSER *new;

    new = newluser();

    strlzcpy(new->u_user, sizeof(new->u_user),
	     ut->ut_name, sizeof(ut->ut_name) );
    strlzcpy(new->u_line, sizeof(new->u_line),
	     ut->ut_line, sizeof(ut->ut_line) );
# ifndef UTMP_NO_HOST
    strlzcpy(new->u_host, sizeof(new->u_host),
	     ut->ut_host, sizeof(ut->ut_host) );
# endif /* UTMP_NO_HOST not defined */
    new->u_time = ut->ut_time;

    return( new );
} /* mkluser */

/*
 * here with tree of usernames from command line
 * return new tree of logged in users
 */

GLOBAL LTREE *ent_select( proto )
LUSER *proto;
{
    register struct utmp *ut;
    char name[ sizeof( ut->ut_name ) + 1 ];
    LUSER *new, *found;
    LTREE *t2;

    t2 = NULL;

    while( (ut = getutent()) != NULL ) {
	if( ut->ut_name[0] == EOS )	/* entry has name? */
	    continue;			/* no, get next */
	if( hideut(ut) )		/* hide login session */
	    continue;
	strlzcpy( name, sizeof(name), ut->ut_name, sizeof( ut->ut_name ) );
	if( (found = treefind( proto, name )) != NULL ) {
	    register PERSON *p;

	    found->u_flags |= U_FOUND;
	    new = mkluser(ut);
	    p = new->u_person = found->u_person;
	    if( p != NULL )
		p->p_count++;
	    t2 = linsert(new, t2);
	} /* found template */
    } /* while utmp */
    endutent();

    return( t2 );
} /* ent_select */

GLOBAL int have_tsels() {
    return first_tsel != NULL;
}

/*
 * Local variables:
 * comment-column: 40
 * End:
 */
