/*
 * kmem.c -- interact with /dev/kmem, /dev/mem etc.
 *
 * Copyright (C) 1986, 1996  Philip L. Budne
 *
 * This file is part of "Phil's Finger Program".
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 */

# ifndef lint
static char *rcsid = "$Id: kmem.c,v 3.8 1999/03/02 04:46:09 phil Exp $";
# endif /* lint not defined */

# include "finger.h"
# ifndef NO_KMEM
# include <sys/types.h>
# include "kmem.h"
# include "args.h"			/* sw_debug */

/* # define DEBUG			/**/

# ifdef sun
# include "info.h"			/* info_[r]consdev */
extern struct info I;			/* namelist values from names.c */

# include <sys/stat.h>			/* to stat /dev/console */
GLOBAL dev_t rconsdev;
# if USG >= 4
# include <fcntl.h>
/* exp; */
GLOBAL dev_t stdindev;
GLOBAL dev_t kbddev;
GLOBAL dev_t fbdev;
# else  /* not USG >= 4 */
GLOBAL dev_t consdev;
# endif /* not USG >= 4 */
GLOBAL dev_t dev_console;
# endif /* sun defined */

GLOBAL FTYPE kmem;			/* fd for /dev/kmem */
# ifndef HAVE_KVM
GLOBAL FTYPE mem;			/* fd for /dev/mem */
# ifdef SWAP_DEVICE
GLOBAL FTYPE drum;			/* fd for /dev/drum */
# endif /* SWAP_DEVICE defined */
# else  /* HAVE_KVM defined */
# include <sys/file.h>			/* O_RDONLY */
# include <stdio.h>			/* NULL */
# endif /* HAVE_KVM defined */

# ifdef NEED_BOOTSEC
# ifdef KSYM_BOOTTIME
# include <sys/time.h>			/* struct timeval */
extern time_t bootsec;
# endif /* KSYM_BOOTTIME defined */
# endif /* NEED_BOOTSEC defined */

LOCAL int ini_kmem2() {
# ifdef sun
    struct stat stb;			/* for /dev/console */
# endif /* sun defined */
# ifdef NEED_BOOTSEC
# ifdef KSYM_BOOTTIME
    struct timeval boottime;
# endif /* KSYM_BOOTTIME defined */
# endif

# ifdef HAVE_KVM
    if( (kmem = kvm_open( NULL, NULL, NULL, O_RDONLY, "finger" )) == NULL ) {
	return( FALSE );
    }
# else  /* HAVE_KVM not defined */
    if( ISBADFILE( (kmem = OPEN("/dev/kmem")) ) ) {
	perror("/dev/kmem");
 	kmem = BADFILE; 	
	return( FALSE );
    }

    if( ISBADFILE( (mem = OPEN("/dev/mem")) ) ) {
	perror("/dev/mem");
	CLOSE(kmem);
	kmem = BADFILE;
	return( FALSE );
    }

# ifdef SWAP_DEVICE
    if( ISBADFILE( (drum = OPEN( SWAP_DEVICE )) ) ) {
	perror(SWAP_DEVICE);
	CLOSE(kmem);
	CLOSE(mem);
 	kmem = BADFILE;
 	return( FALSE );
    }
# endif /* SWAP_DEVICE defined */
# endif /* HAVE_KVM not defined */

# ifndef NO_SYMS
    if( !readnames() )			/* read in info struct */
	return( FALSE );
# endif /* NO_SYMS not defined */

# ifdef sun
    KMEMREAD(I.info_rconsdev, &rconsdev, sizeof(rconsdev));
# if SunOS
    KMEMREAD(I.info_consdev,  &consdev,  sizeof(consdev));
# else  /* not SunOS */
/* exp; */
    KMEMREAD(I.info_stdindev, &stdindev, sizeof(stdindev));
    KMEMREAD(I.info_kbddev, &kbddev, sizeof(kbddev));
    KMEMREAD(I.info_fbdev, &fbdev, sizeof(fbdev));
# endif /* not SunOS */
    if( stat("/dev/console", &stb ) == 0 )
	dev_console = stb.st_rdev;
    else
	dev_console = -1;		/* makedev(-1,-1)? */
# ifdef DEBUGSW
    if( sw_debug ) {
# ifdef SunOS
	printf("consdev %#x ", consdev);
# else  /* SunOS not defined */
/* exp; */
	printf("stdindev %#x kbddev %#x fbdev %#x ", stdindev, kbddev, fbdev );
# endif /* SunOS not defined */
	printf("rconsdev %#x /dev/console %#x\n", rconsdev, dev_console );
    }
# endif /* DEBUGSW defined */
# endif /* sun defined */

# ifdef NEED_BOOTSEC
# ifdef KSYM_BOOTTIME
    if( KMEMREAD(I.info_boottime, &boottime, sizeof(boottime)) ) {
	bootsec = boottime.tv_sec;
	/* WISH: output if sw_debug set? */
    }
# endif /* KSYM_BOOTTIME defined */
# endif /* NEED_BOOTSEC defined */

    return( TRUE );
} /* ini_kmem2 */

GLOBAL int ini_kmem() {
    static int once;
    static int ret;

    if (!once) {
	ret = ini_kmem2();
	once = TRUE;
    }
    return ret;
} /* ini_kmem */

# ifndef HAVE_KVM
GLOBAL int kread(f, pos, buf, count)
FTYPE f;
int count;
unsigned long pos;			/* caddr_t? */
char *buf;
{
    int cc;
# ifdef DEBUG
    char *name;

    if( f == mem )
	name = " mem";
    else if( f == kmem )
	name = "kmem";
    else if( f == drum )
	name = "drum";
    else
	name = "??";
# endif /* DEBUG defined */

    if( SEEK(f, pos, 0) == -1 ) {
        perror("seek");
	cc = -1;
    }
    else 
	cc = READ(f, buf, count);

# if BUFIO
    if( cc > 0 )
	cc *= count;
# endif /* BUFIO */

# ifdef DEBUG
    printf("kread: %s@%08x %d(%d) chars\n",
	   name,
	   pos,
	   count,
	   cc);
# endif /* DEBUG defined */

    if( cc == -1 ) {
# ifdef DEBUG
	perror("read");
# endif /* DEBUG defined */
	return( FALSE );
    } /* bad return */
    return( TRUE );
} /* kread */
# endif /* HAVE_KVM not defined */

# endif /* NO_KMEM not defined */

/*
 * Local variables:
 * comment-column: 40
 * End:
 */
