/*
 * readpr.c -- read process table
 *
 * Copyright (C) 1986, 1996  Philip L. Budne
 *
 * This file is part of "Phil's Finger Program".
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 */

# ifndef lint
static char *rcsid = "$Id: readpr.c,v 3.33 2003/11/08 03:17:44 phil Exp $";
# endif /* lint not defined */

# include "finger.h"
# ifndef NO_WHAT
# include "waitstate.h"
# ifndef Umax

# include <stdio.h>
# include "ustruct.h"

# if defined(__hpux) && !defined(hpux)	/* hpux11 */
# define hpux
# endif /* defined(__hpux) && !defined(hpux) */

# if defined(USG) && USG < 4
# include <sys/var.h>
# endif /* defined(USG) && USG < 4 */

# ifndef SIGHUP
# include <signal.h>
# endif /* SIGHUP not defined */

# include "pr.h"
# include "args.h"			/* for debugsw */

/* # define DEBUG_READPR		/* dump all pr's */

# ifndef UTMP_NO_HOST
# define MUST_HAVE_TTY			/* run a bit faster */
# endif /* UTMP_NO_HOST not defined */

# ifndef NO_SYMS
# include "info.h"			/* after finger.h */
extern struct info I;			/* namelist values from names.c */
# endif /* NO_SYMS not defined */

# ifndef NO_USTRUCT
extern struct user *ustruct();		/* from ustruct.c */
extern void u_cleanup();		/* from ustruct.c */
# endif /* NO_USTRUCT not defined */

# ifndef NO_KMEM
# include "kmem.h"
extern FTYPE kmem;			/* fd for /dev/kmem (or libkvm) */
# ifndef HAVE_KVM
extern FTYPE mem;			/* fd for /dev/mem */
extern FTYPE swap;			/* fd for /dev/drum */
					/* from kmem.c */
# endif /* HAVE_KVM not defined */
# endif /* NO_KMEM not defined */

# ifdef BSD4_4
# include <sys/sysctl.h>
# include <machine/cpu.h>		/* for CPU_CONSDEV */
GLOBAL dev_t consdev = -1;
# endif /* BSD4_4 defined */

FORWARD LOCAL enum sig istate();

LOCAL int nproc;			/* kernel nproc */

/*
 * readpr finds and reads in the array pr, containing the interesting
 * parts of the proc and user tables for each live process.
 */

# ifdef AIX_RS6K
# define HAVE_READPR
# define MAXPROC 512

GLOBAL struct pr *readpr() {
    int i, n;
    struct pr *prvec;
    struct procinfo pivec[MAXPROC];
    register struct pr *prp;
    register struct procinfo *pip;

    n = getproc( pivec, MAXPROC, sizeof( struct procinfo ) );
    if( n < 1 ) {
        /* Check errno, report of MAXPROC too small!!! */
	return( NULL );
    }

    prvec = (struct pr *) calloc( n+1, sizeof( struct pr ) );
    if( prvec == NULL )
	return( NULL );

    prp = prvec;
    for( i = 0, pip = pivec; i < n; i++, pip++ ) {
	struct userinfo ui;

	if( pip->pi_stat == 0 )
	    continue;

	if( getuser( pip, sizeof( struct procinfo ),
		    &ui,  sizeof( struct userinfo ) ) < 0 )
	    continue;

	prp->pr_uid = pip->pi_uid;
	prp->pr_pid = pip->pi_pid;
	prp->pr_ppid = pip->pi_ppid;
	prp->pr_pgrp = pip->pi_pgrp;
	prp->pr_stat = pip->pi_stat;
	prp->pr_flag = pip->pi_flag;
	prp->pr_wchan = pip->pi_wchan;
	prp->pr_ttyp = ui.ui_ttyp;
	prp->pr_ttyd = ui.ui_ttyd + ui.ui_ttympx;

	strlzcpy(prp->pr_cmd, sizeof(prp->pr_cmd),
		 ui.ui_comm, sizeof( ui.ui_comm ) );
# ifdef DEBUG_READPR
	dumppr( prp );
# endif /* DEBUG_READPR defined */
	prp++;
    }
    prp->pr_stat = 0;			/* mark end */

# ifdef DEBUGSW
    if( sw_debug )
	printf("read %d processes\n", prp - prvec );
# endif /* DEBUGSW defined */

    /* realloc prvec to size? */
    return( prvec );
}
# endif /* AIX_RS6K defined */

# ifdef SVR4_PROCFS
# define HAVE_READPR

# include <dirent.h>
# include <sys/procfs.h>

GLOBAL struct pr *readpr() {
    DIR *dir;
    struct dirent *dp;
    struct pr *prvec;
    register struct pr *prp;		/* current pr entry */
    int maxproc;

# define MAXPROC 1024			/* get from v.v_proc??? */
    maxproc = MAXPROC;
    prvec = (struct pr *)calloc(maxproc+1, sizeof(struct pr));
    if( prvec == NULL )
	return( NULL );

    dir = opendir("/proc");
    if( dir == NULL )
	return( NULL );

    nproc = 0;
    prp = prvec;
    while ((dp = readdir(dir)) != NULL && nproc < maxproc) {
# ifdef PIOCGETU
	struct user u;
# endif /* PIOCGETU defined */
	struct prpsinfo prps;
	char path[32];
	int f;

	if (dp->d_name[0] == '.')
	    continue;

	sprintf( path, "/proc/%s", dp->d_name );
	f = open( path, 0 );
	if (f < 0)
	    continue;

	if (ioctl( f, PIOCPSINFO, &prps) < 0) {	/* get "ps" info */
	    close(f);
	    continue;
	}

	if( prps.pr_zomb || prps.pr_sname == 'Z') { /* zombie */
	    close(f);
	    continue;
	}

	if (prps.pr_sname == 'I') {	/* idle: undergoing creation */
	    close(f);
	    continue;
	}

# ifdef SSYS
	if( prps.pr_flag & SSYS ) {	/* swapper or pager? */
	    close(f);
	    continue;			/* flush */
	}
# endif /* SSYS defined */

	/* save the interesting parts */
	prp->pr_uid = prps.pr_uid;	/* user id */
	prp->pr_pid = prps.pr_pid;	/* pid */
	prp->pr_ppid = prps.pr_ppid;	/* parent pid */
	prp->pr_pgrp = prps.pr_pgrp;	/* process group leader pid */
	prp->pr_flag = prps.pr_flag;	/* flags (esp swap) */
	prp->pr_stat = prps.pr_state;	/* state (1..6) */
	prp->pr_wchan = prps.pr_wchan;	/* wait condition */
	prp->pr_ttyd = prps.pr_ttydev;	/* device number */
	strlzcpy(prp->pr_cmd, sizeof(prp->pr_cmd),
		 prps.pr_fname, sizeof(prps.pr_fname));

# ifdef PIOCGETU
	/* crushes other local variables?? */
	if (ioctl(f, PIOCGETU, &u) == 0) {
	    prp->pr_ttyp = u.u_ttyp;	/* pointer to tpg (always NULL?) */
# define NEED_ISTATE
	    prp->pr_hup  = istate( u.u_signal[SIGHUP] );
	    prp->pr_intr = istate( u.u_signal[SIGINT] );
	    prp->pr_quit = istate( u.u_signal[SIGQUIT]);
	}
# endif /* PIOCGETU defined */
	close(f);

# ifdef DEBUG_READPR
	dumppr( prp );
# endif /* DEBUG_READPR defined */
	prp++;
	nproc++;
	if (nproc == maxproc) {
	    maxproc <<= 1;		/* double it! */
	    prvec = (struct pr *)realloc(prvec,
					 (maxproc+1)*sizeof(struct pr));
	    prp = prvec + nproc;
	    /* check for failure? do what??? */
	}
    }
    prp->pr_stat = 0;			/* mark end */
    
# ifdef DEBUGSW
    if( sw_debug )
	printf("read %d processes\n", nproc);
# endif /* DEBUGSW defined */

    return( prvec );
}
# endif /* SVR4_PROCFS defined */

# ifdef BSD4_4
# define HAVE_READPR

# ifdef BSD_PROCFS
# if defined(__bsdi__) && defined(d_type)
/*
 * On BSDI 2.1 d_type is #defined in <machine/segments.h>, included by
 * <machine/cpu.h>.  This conflicts with the d_type used in struct dirent
 * 
 * spike@indra.com 7/18/96
 */
# undef d_type
# endif /* defined(__bsdi__) && defined(d_type) */
# include <dirent.h>
# else  /* BSD_PROCFS not defined */
# include <sys/user.h>
# endif /* BSD_PROCFS not defined */

GLOBAL struct pr *readpr() {
/* use /proc for starters; doesn't give state/flags/signals */
# ifdef BSD_PROCFS
    DIR *dir;
    struct dirent *dp;
# else  /* BSD_PROCFS not defined */
    struct kinfo_proc *kp;
    int i;
# endif /* BSD_PROCFS not defined */
    int maxproc;
    struct pr *prvec;
    register struct pr *prp;		/* current pr entry */
    int mib[2];
    size_t len;
# ifdef NEED_BOOTSEC
    struct timeval boottime;
    extern time_t bootsec;
# endif /* NEED_BOOTSEC defined */

# ifndef NO_KMEM
    if( !ini_kmem() )
	return( NULL );
# endif /* NO_KMEM not defined */

# ifdef CPU_CONSDEV
    mib[0] = CTL_MACHDEP;
    mib[1] = CPU_CONSDEV;
    len = sizeof(consdev);
    sysctl(mib, 2, &consdev, &len, NULL, 0);
# ifdef DEBUGSW
    if( sw_debug )
	printf("consdev %#x\n", consdev);
# endif /* DEBUGSW defined */
# endif /* CPU_CONSDEV defined */

# ifdef NEED_BOOTSEC
    mib[0] = CTL_KERN;
    mib[1] = KERN_BOOTTIME;
    len = sizeof(boottime);
    if( sysctl(mib, 2, &boottime, &len, NULL, 0) == 0 )
	bootsec = boottime.tv_sec;
# endif /* NEED_BOOTSEC defined */

# ifdef BSD_PROCFS
    mib[0] = CTL_KERN;
    mib[1] = KERN_MAXPROC;
    len = sizeof(maxproc);
    if( sysctl(mib, 2, &maxproc, &len, NULL, 0) < 0 )
	return NULL;

    prvec = (struct pr *)calloc(maxproc+1, sizeof(struct pr));
    if( prvec == NULL )
	return( NULL );

    dir = opendir("/proc");
    if( dir == NULL ) {
	free(prvec);
	return( NULL );
    }

    nproc = 0;
    prp = prvec;
    while ((dp = readdir(dir)) != NULL && nproc < maxproc) {
	char path[64], data[129], flags[32], command[32], wmesg[32];
	int pid, ppid, pgrp, sid, uid, gid;
	int ttymaj,ttymin;
	int ssec, susec;		/* start time */
	int usec, uusec;		/* user time */
	int sysec, syusec;		/* system time */
	int euid, egid;
	int f, cc;

	if (dp->d_name[0] == '.' ||	/* skip . and .. */
	    dp->d_name[0] == 'c' ||	/* skip curproc */
	    dp->d_name[0] == 's')	/* skip self (NetBSD) */
	    continue;

	sprintf( path, "/proc/%s/status", dp->d_name );
	f = open( path, 0 );
	if (f < 0)
	    continue;

	cc = read(f, data, sizeof(data));
	/* XXX fstat()??? */
	close(f);
	if (cc <= 0)
	    continue;
	data[cc] = '\0';

	sscanf( data, "%s%d%d%d%d%d,%d%s%d,%d%d,%d%d,%d%s%d%d",
	       command, &pid, &ppid, &pgrp, &sid, &ttymaj, &ttymin,
	       flags, &ssec, &susec, &usec, &uusec, &sysec, &syusec, wmesg,
	       &euid, &egid );

	/* save the interesting parts */
	prp->pr_uid = euid;		/* sigh; _effective_ user id */
	prp->pr_pid = pid;		/* pid */
	prp->pr_ppid = ppid;		/* parent pid */
	prp->pr_pgrp = pgrp;		/* process group leader pid */
	prp->pr_ttyd = makedev(ttymaj,ttymin); /* device number */
	prp->pr_stat = SSLEEP;		/* UGH: must be non-zero */

	strlzcpy(prp->pr_cmd, sizeof(prp->pr_cmd), command, sizeof(command));
	strlzcpy(prp->pr_wmesg, sizeof(prp->pr_wmesg), wmesg, sizeof(wmesg));

# ifdef DEBUG_READPR
	dumppr( prp );
# endif /* DEBUG_READPR defined */
	prp++;
	nproc++;
    } /* readdir */
# else  /* BSD_PROCFS not defined */

    /* requires /dev/kmem access, but returns more complete info */

    kp = kvm_getprocs(kmem, KERN_PROC_ALL, 0, &maxproc);
    if (kp == NULL)
	return NULL;

    prvec = (struct pr *)calloc(maxproc+1, sizeof(struct pr));
    if( prvec == NULL )
	return( NULL );

    nproc = 0;
    prp = prvec;
    for (i = 0; i < maxproc; i++, kp++) {

# ifdef FREEBSD5_KINFO_PROC
	if( kp->ki_stat == 0 || kp->ki_stat == SZOMB ) /* a zombie or worse? */
	    continue;			/* forget it */

	if( kp->ki_flag & (P_SYSTEM|P_WEXIT)) /* system or exiting? */
	    continue;			/* nevermind. */

	prp->pr_uid = kp->ki_ruid;	/* real uid */
	prp->pr_pid = kp->ki_pid;	/* pid */
	prp->pr_ppid = kp->ki_ppid;	/* parent pid */
	prp->pr_pgrp = kp->ki_pgid;	/* process group leader pid */
	prp->pr_ttyd = kp->ki_tdev;	/* device number */
	prp->pr_stat = kp->ki_stat;	/* state (1..6) */
	prp->pr_flag = kp->ki_flag;	/* flags (esp swap) */

	strlzcpy(prp->pr_cmd, sizeof(prp->pr_cmd),
		 kp->ki_comm, sizeof(kp->ki_comm));
	strlzcpy(prp->pr_wmesg, sizeof(prp->pr_wmesg),
		 kp->ki_wmesg, sizeof(kp->ki_wmesg));

# ifdef SIGISMEMBER
# define SIGSTATE(SIGN) \
	(SIGISMEMBER(kp->ki_sigignore,SIGN) ? S_IGNORE : \
	 (SIGISMEMBER(kp->ki_sigcatch,SIGN) ? S_CATCH : S_DEFAULT))

	prp->pr_hup  = SIGSTATE(SIGHUP);
	prp->pr_intr = SIGSTATE(SIGINT);
	prp->pr_quit = SIGSTATE(SIGQUIT);
# endif /* SIGISMEMBER defined */
# else  /* FREEBSD5_KINFO_PROC not defined */
	struct proc *mpp;
	struct eproc *epp;

	mpp = &kp->kp_proc;

	if( mpp->p_stat == 0 || mpp->p_stat == SZOMB ) /* a zombie or worse? */
	    continue;			/* forget it */

# ifdef P_SYSTEM
	if( mpp->p_flag & P_SYSTEM )	/* swapper or pager? */
	    continue;			/* flush */
# endif /* P_SYSTEM defined */
# ifdef P_WEXIT
	if( mpp->p_flag & P_WEXIT )	/* exiting? */
	    continue;			/* nevermind. */
# endif /* P_WEXIT defined */
# ifdef SSYS
	if( mpp->p_flag & SSYS )	/* swapper or pager? */
	    continue;			/* flush */
# endif /* SSYS defined */
# ifdef SWEXIT
	if( mpp->p_flag & SWEXIT )	/* exiting? */
	    continue;			/* nevermind. */
# endif /* SWEXIT defined */

	epp = &kp->kp_eproc;
	
	prp->pr_uid = epp->e_pcred.p_ruid; /* real uid */
	prp->pr_pid = mpp->p_pid;	/* pid */
	prp->pr_ppid = epp->e_ppid;	/* parent pid */
	prp->pr_pgrp = epp->e_pgid;	/* process group leader pid */
	prp->pr_ttyd = epp->e_tdev;	/* device number */
	prp->pr_stat = mpp->p_stat;	/* state (1..6) */
	prp->pr_flag = mpp->p_flag;	/* flags (esp swap) */

	strlzcpy(prp->pr_cmd, sizeof(prp->pr_cmd),
		 mpp->p_comm, sizeof(mpp->p_comm));
	strlzcpy(prp->pr_wmesg, sizeof(prp->pr_wmesg),
		 epp->e_wmesg, sizeof(wpp->e_wmesg));

# ifdef SIGISMEMBER		/* FreeBSD; on NetBSD need kinfo_proc2? */
# define SIGSTATE(SIGN) \
	(SIGISMEMBER(epp->e_procsig.ps_sigignore,SIGN) ? S_IGNORE : \
	 (SIGISMEMBER(epp->e_procsig.ps_sigcatch,SIGN) ? S_CATCH : S_DEFAULT))

	prp->pr_hup  = SIGSTATE(SIGHUP);
	prp->pr_intr = SIGSTATE(SIGINT);
	prp->pr_quit = SIGSTATE(SIGQUIT);
# endif /* SIGISMEMBER defined */
# endif /* FREEBSD5_KINFO_PROC not defined */

# ifdef DEBUG_READPR
	dumppr( prp );
# endif /* DEBUG_READPR defined */
	prp++;
	nproc++;
    }
# endif /* BSD_PROCFS not defined */
    prp->pr_stat = 0;			/* mark end */
# ifdef DEBUGSW
    if( sw_debug )
	printf("read %d processes\n", prp - prvec );
# endif /* DEBUGSW defined */
    return( prvec );
}
# endif /* BSD4_4 defined */

# ifdef HPUX_PSTAT
# define HAVE_READPR

# include <sys/pstat.h>

GLOBAL dev_t consdev = -1;
# ifdef NEED_BOOTSEC
extern time_t bootsec;
# endif /* NEED_BOOTSEC defined */

GLOBAL struct pr *readpr() {
    struct pst_static stat;
    struct pst_status *procs;
    union pstun pstun;
    int maxproc, nproc;
    struct pr *prvec;
    register struct pr *prp;		/* current pr entry */
    register struct pst_status *psp;

    if (!readnames())
	return NULL;

    /* use old interface for compatibility */
    pstun.pst_static = &stat;
    if (pstat(PSTAT_STATIC, pstun, sizeof(stat), 0, 0) != 1)
	return NULL;

    maxproc = stat.max_proc;
    consdev = makedev(stat.console_device.psd_major,
		      stat.console_device.psd_minor);
# ifdef NEED_BOOTSEC
    bootsec = stat.boot_time;
# endif /* NEED_BOOTSEC defined */
# ifdef DEBUGSW
    if( sw_debug )
	printf("consdev %#x\n", consdev);
# endif /* DEBUGSW defined */

    procs = (struct pst_status *)calloc(maxproc, sizeof(struct pst_status));
    if (!procs)
	return NULL;

    pstun.pst_status = procs;
    nproc = pstat(PSTAT_PROC, pstun, sizeof(struct pst_status), maxproc, 0);
    if (nproc < 1) {
	free(procs);
	return NULL;
    }

    prvec = (struct pr *)calloc(nproc+1, sizeof(struct pr));
    if (!prvec) {
	free(procs);
	return NULL;
    }

    for (psp = procs, prp = prvec; nproc > 0; psp++, nproc--) {
	if (psp->pst_stat == 0 || psp->pst_stat == PS_ZOMBIE)
	    continue;			/* skip zombies or worse */

	/* XXX skip if stat PS_IDLE or PS_OTHER? */

	if (psp->pst_flag & PS_SYS)
	    continue;			/* skip system processes */

	prp->pr_uid = psp->pst_uid;	/* uid */
	prp->pr_pid = psp->pst_pid;	/* pid */
	prp->pr_ppid = psp->pst_ppid;	/* parent */
	prp->pr_pgrp = psp->pst_pgrp;	/* process group leader */
	prp->pr_ttyd = makedev(psp->pst_major, psp->pst_minor); /* tty dev_t */
	prp->pr_stat = psp->pst_stat;	/* process state */
	prp->pr_flag = psp->pst_flag;	/* flags (esp swap) */
	prp->pr_wchan = psp->pst_wchan;	/* wait condition */

	strlzcpy(prp->pr_cmd, sizeof(prp->pr_cmd),
		 psp->pst_ucomm, sizeof(psp->pst_ucomm));

# ifdef DEBUG_READPR
	dumppr( prp );
# endif /* DEBUG_READPR defined */
	prp++;
    } /* for each proc */

# ifdef DEBUGSW
    if( sw_debug )
	printf("read %d processes\n", prp - prvec );
# endif /* DEBUGSW defined */

    free(procs);
    return prvec;
}
# endif /* HPUX_PSTAT defined */

# ifdef SOLARIS_PROCFS
# define HAVE_READPR

# include <dirent.h>
# include <procfs.h>

GLOBAL struct pr *readpr() {
    DIR *dir;
    struct dirent *dp;
    struct pr *prvec;
    register struct pr *prp;		/* current pr entry */
    int maxproc;
    uid_t myuid;

# ifndef NO_KMEM
    if( !ini_kmem() )
	return( NULL );
# endif /* NO_KMEM not defined */

# define MAXPROC 1024			/* get from v.v_proc??? */
    maxproc = MAXPROC;
    prvec = (struct pr *)calloc(maxproc+1, sizeof(struct pr));
    if( prvec == NULL )
	return( NULL );

    dir = opendir("/proc");
    if( dir == NULL )
	return( NULL );

    nproc = 0;
    prp = prvec;
    myuid = getuid();
    while ((dp = readdir(dir)) != NULL && nproc < maxproc) {
	struct psinfo ps;
	char path[64];
	int f;

	if (dp->d_name[0] == '.')
	    continue;
	sprintf( path, "/proc/%s/psinfo", dp->d_name );
	f = open( path, 0 );
	if (f < 0)
	    continue;

	if (read( f, &ps, sizeof(ps)) < sizeof(ps)) {
	    close(f);
	    continue;
	}
	close(f);

	if( ps.pr_lwp.pr_sname == 'Z')	/* zombie */
	    continue;

	if (ps.pr_lwp.pr_sname == 'I')	/* idle: undergoing creation */
	    continue;

# ifdef SSYS
	if( ps.pr_flag & SSYS )		/* swapper or pager? */
	    continue;			/* flush */
# endif /* SSYS defined */

	    /* save the interesting parts */
	prp->pr_uid = ps.pr_uid;	/* user id */
	prp->pr_pid = ps.pr_pid;	/* pid */
	prp->pr_ppid = ps.pr_ppid;	/* parent pid */
	prp->pr_pgrp = ps.pr_pgid;	/* process group leader pid */
	prp->pr_flag = ps.pr_flag;	/* flags (esp swap) */
	prp->pr_stat = ps.pr_lwp.pr_state; /* state (1..6) */
	prp->pr_wchan = (caddr_t) ps.pr_lwp.pr_wchan; /* wait condition */
	prp->pr_ttyd = ps.pr_ttydev;	/* device number */
	strlzcpy(prp->pr_cmd, sizeof(prp->pr_cmd),
		 ps.pr_fname, sizeof(ps.pr_fname));

	/* only try reading sigactions if we're superuser, or same uid */
	if (myuid == 0 || myuid == ps.pr_uid) {
	    sprintf( path, "/proc/%s/sigact", dp->d_name );
	    f = open( path, 0 );
	    if (f >= 0) {
		struct sigaction sig[SIGQUIT];	/* XXX highest sig we want */

		if (read(f, &sig, sizeof(sig)) == sizeof(sig)) {
# define NEED_ISTATE
		    prp->pr_hup  = istate( sig[SIGHUP-1].sa_handler );
		    prp->pr_intr = istate( sig[SIGINT-1].sa_handler );
		    prp->pr_quit = istate( sig[SIGQUIT-1].sa_handler );
		}
		close(f);
	    }
	}

# ifdef DEBUG_READPR
	dumppr( prp );
# endif /* DEBUG_READPR defined */
	prp++;
	nproc++;
	if (nproc == maxproc) {
	    maxproc <<= 1;		/* double it! */
	    prvec = (struct pr *)realloc(prvec,
					 (maxproc+1)*sizeof(struct pr));
	    /* check for failure? do what??? */
	    prp = prvec + nproc;
	}
    }
    prp->pr_stat = 0;			/* mark end */
    
# ifdef DEBUGSW
    if( sw_debug )
	printf("read %d processes\n", nproc );
# endif /* DEBUGSW defined */

    return( prvec );
}
# endif /* SOLARIS_PROCFS defined */

/* "generic" version */
# ifndef HAVE_READPR
# define HAVE_APROC
LOCAL struct proc *aproc;		/* kernel virt addr of process table */

GLOBAL struct pr *readpr() {
    register struct pr *prp;		/* current pr entry */
    register int i;			/* loop index */
    struct user *upp;			/* pointer to upage(s) */
    struct proc *mpp;			/* ptr to current process */
    struct pr *prvec;
    struct proc *allproc;		/* whole process table */
# ifdef USG
    struct var v;
# endif /* USG defined */
# if SunOS >= 410
    struct sess sess;
# endif /* SunOS >= 410 */

    if( !ini_kmem() )			/* initialize /dev/kmem etc. */
	return( NULL );

# ifdef USG
    KMEMREAD(I.info_v, &v, sizeof( v ) ); /* var struct */
    nproc = v.v_proc;			/* size of proc table */
    aproc = (struct proc *) I.info_proc; /* addr of proc table */
# else  /* USG not defined */
    KMEMREAD(I.info_nproc, &nproc, sizeof( nproc )); /* size of proc tbl */
    KMEMREAD(I.info_proc,  &aproc, sizeof( aproc )); /* addr of proc tbl */
# endif /* USG not defined */

# ifdef DEBUG_READPR
    printf("nproc %d aproc %#x\n", nproc, aproc );
# endif /* DEBUG_READPR defined */

    prvec = (struct pr *)calloc(nproc+1, sizeof(struct pr));/* allocate tbl */
    if( prvec == NULL )
	return( NULL );

    allproc = (struct proc *) calloc(nproc, sizeof( struct proc ) );
    if( allproc == NULL ) {
	free(prvec);
	prvec = NULL;
	return( NULL );
    }
    KMEMREAD((int)aproc, allproc, nproc * sizeof( struct proc ) );

    mpp = allproc;
    prp = prvec;			/* pointer to dest block */
    for( i = 0; i < nproc; i++, mpp++ ) { /* for all processes */
	if( mpp->p_stat == 0 || mpp->p_stat == SZOMB ) /* a zombie or worse? */
	    continue;			/* forget it */

# ifdef ULTRIX4
# define p_flag p_type
# endif /* ULTRIX4 defined */

# ifdef SSYS
	if( mpp->p_flag & SSYS )	/* swapper or pager? */
	    continue;			/* flush */
# endif /* SSYS defined */

# ifdef SWEXIT
	if( mpp->p_flag & SWEXIT )	/* exiting? */
	    continue;			/* nevermind. */
# endif /* SWEXIT defined */

# if defined(MUST_HAVE_TTY) && defined(P_TTYP)
	/* avoid fetching ustruct needlessly */
	if( mpp->p_ttyp == NULL )	/* no terminal pointer? */
	    continue;
# endif /* defined(MUST_HAVE_TTY) && defined(P_TTYP) */

# if SunOS >= 410
	if( mpp->p_sessp == NULL ||
	   !KMEMREAD( (long)mpp->p_sessp, (char *)&sess, sizeof(sess) ) ) {
# ifdef MUST_HAVE_TTY
	    continue;
# else  /* MUST_HAVE_TTY not defined */
	    sess.s_ttyp = NULL;
# endif /* MUST_HAVE_TTY not defined */
	} /* no session or kmem read failed */
# endif /* SunOS >= 410 */

	if( (upp = ustruct( mpp )) == NULL ) /* get ustruct? */
	    continue;			/* no point. */

# if defined(DEBUGSW) && USG < 4 && !defined(hpux)
	if( sw_debug && ((unsigned)upp->u_procp) != ((unsigned)(aproc + i)) )
	   printf("%%Funny u struct for pid %d u_procp %#x != %#x\n",
		  mpp->p_pid, upp->u_procp, (aproc + i) );
# endif /* defined(DEBUGSW) && USG < 4 && !defined(hpux) */

	/*
	 * WISH:
	 * check for master /etc/inetd /etc/rlogind etc???  (as a way to
	 * establish root of a process tree and avoid stray background
	 * processes on that tty) and keep only descendants thereof.
	 */

# if 0 /* since we already have ustruct, just keep it! */
# if defined(MUST_HAVE_TTY) && !defined(P_TTYP) && SunOS < 410
	if( upp->u_ttyp == NULL )	/* no terminal pointer? */
	    continue;
# endif /* defined(MUST_HAVE_TTY) && !defined(P_TTYP) && SunOS < 410 */
# endif /* 0 */

	/* save the interesting parts */
# ifdef ULTRIX4
# undef p_flag
# define p_flag p_sched			/* for SLOAD flag */
# endif /* ULTRIX4 defined */

	prp->pr_uid = mpp->p_uid;	/* user id */
	prp->pr_pid = mpp->p_pid;	/* pid */
	prp->pr_ppid = mpp->p_ppid;	/* parent pid */
	prp->pr_pgrp = mpp->p_pgrp;	/* process group leader pid */
	prp->pr_stat = mpp->p_stat;	/* state (1..6) */
	prp->pr_flag = mpp->p_flag;	/* flags (esp swap) */
	prp->pr_wchan = mpp->p_wchan;	/* wait address */

# if SunOS >= 410
	prp->pr_ttyp = sess.s_ttyp;
	prp->pr_ttyd = sess.s_ttyd;
# else  /* not SunOS >= 410 */
# ifdef P_TTYP
	prp->pr_ttyp = mpp->p_ttyp;	/* pointer to tty structure */
# else  /* P_TTYP not defined */
	prp->pr_ttyp = upp->u_ttyp;	/* pointer to tty structure or tpg */
# endif /* P_TTYP not defined */
# ifdef P_TTYD
	prp->pr_ttyd = mpp->p_ttyd;	/* device number */
# else  /* P_TTYD not defined */
	prp->pr_ttyd = upp->u_ttyd;	/* device number */
# endif /* P_TTYD not defined */
# endif /* not SunOS >= 410 */

# define NEED_ISTATE
	prp->pr_hup  = istate( upp->u_signal[SIGHUP] );
	prp->pr_intr = istate( upp->u_signal[SIGINT] );
	prp->pr_quit = istate( upp->u_signal[SIGQUIT]);

	/* what you've all been waiting for... */
	strlzcpy(prp->pr_cmd, sizeof(prp->pr_cmd),
		 upp->u_comm, sizeof(upp->u_comm) );

# ifdef DEBUG_READPR
	dumppr( prp );
# endif /* DEBUG_READPR defined */
	prp++;
    } /* for i */
    prp->pr_stat = 0;			/* mark end */
    
    /* WISH: realloc to shorten prvec? */

    free( allproc );			/* free copy of process table */
    u_cleanup();			/* cleanup ustruct reader */

# ifdef DEBUGSW
    if( sw_debug )
	printf("read %d processes\n", prp - prvec );
# endif /* DEBUGSW defined */

    return( prvec );
} /* readpr */
# endif /* HAVE_READPR not defined */

# ifdef NEED_ISTATE
LOCAL enum sig istate( x )		/* return signal handler state */
# ifdef VOIDSIG
    void (*x)();
# else  /* VOIDSIG not defined */
    int (*x)();
# endif /* VOIDSIG not defined */
{
    if( x == SIG_DFL )
        return( S_DEFAULT );		/* no handler */
    else if( x == SIG_IGN )
        return( S_IGNORE );		/* ignores interrupts */
# ifdef SIG_HOLD
    else if( x == SIG_HOLD )
        return( S_HOLD );		/* interrupt held? */
# endif /* SIG_HOLD defined */
    else
	return( S_CATCH );		/* has handler */
} /* istate */
# endif /* NEED_ISTATE defined */

GLOBAL waitstate_t waitstate( prp )	/* look at process wait states */
    struct pr *prp;
{
# ifdef BSD4_4
# define CMP(S) (strcmp(prp->pr_wmesg, S) == 0)
    if (CMP("nochan"))
	return WS_RU;
    if (CMP("ttyin"))
	return WS_TI;
    if (CMP("ttyout"))
	return WS_TO;
    if (strncmp(prp->pr_wmesg, "tty", 3) == 0)
	return WS_TW;
    if (CMP("select"))
	return WS_SE;
    if (CMP("pause"))
	return WS_PA;
    if (CMP("wait"))
	return WS_WA;
# else  /* BSD4_4 not defined */
    register caddr_t w = prp->pr_wchan;

# if !defined(NO_TTYP) && defined(HAVE_SYS_TTY_H) && !defined(SHORT_TTYP)
    register struct tty *t;		/* non-streams ttys */

    t = prp->pr_ttyp;			/* get tty pointer */

    if( w == (caddr_t) &t->t_rawq )	/* input wait */
	return( WS_TI );
    if( w == (caddr_t) &t->t_outq )	/* output wait */
	return( WS_TO );
    if( w >= (caddr_t) t && w < (caddr_t) (t+1) ) /* other terminal wait */
	return( WS_TW );
# endif /* !defined(NO_TTYP) && defined(HAVE_SYS_TTY_H) && !defined(SHORT_TTYP) */

    if( w == (caddr_t) 0 )		/* no wait? */
	return( WS_RU );		/* runable */

# ifndef NO_SYMS
# ifdef USG
# if defined(STREAMS) && USG < 4
    if( w == (caddr_t) I.info_pollwait ) /* poll(2) */
	return( WS_SE );
# endif /* defined(STREAMS) && USG < 4 */
# else  /* USG not defined */
    if( w == (caddr_t) I.info_selwait )	/* select(2) */
	return( WS_SE );
# endif /* USG not defined */

# if USG < 4 && !defined(hpux)
# if SunOS < 410
    if( w == (caddr_t) I.info_u )	/* pause(2) */
	return( WS_PA );
# endif /* SunOS < 410 */
# endif /* USG < 4 && !defined(hpux) */

# endif /* NO_SYMS not defined */

# ifdef HAVE_APROC
    /* proc's are chained in SVR4! */
    if( w >= (caddr_t) aproc && w <= (caddr_t) (aproc+nproc) ) /* wait(2)*/
	return( WS_WA );
# endif /* HAVE_APROC defined */
# endif /* BSD4_4 not defined */
    return( WS_SL );			/* something else */
} /* waitstate */
# endif /* Umax not defined */
# endif /* NO_WHAT not defined */

/*
 * Local variables:
 * comment-column: 40
 * End:
 */
