/*
 * safe.c -- safe file open routines
 *
 * Copyright (C) 1996  Philip L. Budne
 *
 * This file is part of "Phil's Finger Program".
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 */

# include <sys/types.h>
# include <sys/stat.h>
# include <stdio.h>
# include "finger.h"
# include "person.h"

# ifdef STRICT_ACCESS
# include <sys/file.h>
# ifndef R_OK
# define R_OK 4
# endif /* R_OK not defined */
# endif /* STRICT_ACCESS defined */

/*
 * created from seperate checks in code from bzs@world.std.com;
 * allows finger to be run safely suid root (the world's home directories
 * are protected) to get file access, but avoid abuse of symlinks!!
 */

/*
 * "safe" opens are used for files in other user's home dir's;
 * 	.forward, .plan, .project
 */

/* tempting to write as return (xxx && yyy && zzz); but hard to single step! */
/* WISH: have wrapper function that syslog()'s failures?! */

GLOBAL BOOL safe(fname, pp)
    char *fname;
    PERSON *pp;
{
    struct stat st;
# ifdef STRICT_ACCESS
    /*
     * allow access only if the person running finger (real UID/GID)
     * could read the file themselves.  The entire path leading to
     * the file must allow access (ie; the home directory must allow
     * at LEAST search (execute) permission).
     */
    if (access(fname, R_OK) < 0)
	return FALSE;
# endif /* STRICT_ACCESS defined */

    if (lstat(fname,&st) < 0)
	return FALSE;

    if ((st.st_mode & S_IFMT) != S_IFREG) /* plain file; reject symlinks */
	return FALSE;

    if (st.st_uid != pp->p_uid)		/* owned by person being finger'ed */
	return FALSE;

    /* and readable (by owner or other?!) */
    if ((st.st_mode & (S_IREAD|S_IROTH)) == 0)
	return FALSE;

    return TRUE;
} /* safe */

GLOBAL int safe_open(fname, pp)
    char *fname;
    PERSON *pp;
{
    if (!safe(fname,pp))
	return -1;
    return open(fname, 0);
}

GLOBAL FILE *safe_fopen(fname, pp)
    char *fname;
    PERSON *pp;
{
    if (!safe(fname,pp))
	return NULL;
    return fopen(fname, "r");
}

/*
 * Local variables:
 * comment-column: 40
 * End:
 */
